#!/usr/bin/perl -w
#
# Script for comparing po-files ...
#
# Parameters: 1st parameter   - first  po-file
#             2nd parameter   - second po-file
#
# 2020.05.11, v. 0.01 - 1-st version
#
# Copyright 2020 Serhij Dubyk (sergijdubyk@gmail.com)
#
# Script 'po_compare.pl' is free software; you can redistribute
# it and/or modify it under the terms of the GNU General Public License 
# as published by the Free Software Foundation; either version 2 
# of the License, or (at your option) any later version.

my $start_time          = time();
my $previous_point_time = time();

use Encode;
use utf8;
use strict;
use Locale::PO;
use Switch;
use Data::Dumper;
#use Locale::Language;
use POSIX;
use Locale::gettext;
use Text::Diff::FormattedHTML;


#setlocale(LC_MESSAGES, "uk_UA");

my $encoding_terminal_output = "utf-8";
my $file_system_encoding     = "utf-8";
my $encoding_input_file      = "utf-8"; # may be changed to charset from header
my $encoding_output_file     = "utf-8";

my $d = Locale::gettext->domain('po_compare');
$d->dir('po/');

#&print_terminal($d->get("Hello world")."\n");

# ------------------- print info -----------------------------------------------------------------------------------

&print_terminal($d->get("Script 'po_compare.pl' © Serhij Dubyk 2020\n"));
&print_terminal($d->get("Arguments: 1st parameter   - first  po-file,\n"));
&print_terminal($d->get("           2nd parameter   - second po-file,\n"));
&print_terminal($d->get("           3rd parameter   - URL of first po-file in Pootle,\n"));
&print_terminal($d->get("           4th parameter   - URL of second po-file in Pootle.\n"));
&print_terminal("\n");

# --------------------- read arguments -----------------------------------------------------------------------------

my @args;
if($ENV{'QUERY_STRING'})
   {@args = split(/\?/,$ENV{'QUERY_STRING'});}
else
   {@args = @ARGV;}
my $po_file1 = $args[0];
my $po_file2 = $args[1];
my $po_url1  = $args[2];
my $po_url2  = $args[3];

&print_terminal($d->get("po-file1 = ")."'$po_file1'\n");
&print_terminal($d->get("po-file2 = ")."'$po_file2'\n");
&print_terminal($d->get("po-url1 = ")."'$po_url1'\n");
&print_terminal($d->get("po-url2 = ")."'$po_url2'\n");

# -------------------------------- load po-file1 ------------------------------------------------------------
&print_terminal($d->get("\n============== Read file '$po_file1':"));
my $po1 = &po_parse($po_file1);

# -------------------------------- load po-file2 ------------------------------------------------------------
&print_terminal($d->get("\n============== Read file '$po_file2':"));
my $po2 = &po_parse($po_file2);

#
my @pies1 = split(/\//,$po_file1);
my @pies2 = split(/\//,$po_file2);

my $csv_file = $pies1[-1]."_AND_".$pies2[-1].".csv";
open(my $CSV, ">",&encode($file_system_encoding,$csv_file)) or die "Can't open file ".&print_terminal($csv_file),": $!";


my $html_file = $pies1[-1]."_AND_".$pies2[-1].".html";
open(my $FILE, ">",&encode($file_system_encoding,$html_file)) or die "Can't open file ".&print_terminal($html_file),": $!";

my $current_date = strftime "%Y%m%d", localtime;
&print_file($CSV,"$current_date\t");

my @file_pies = split(/\//,$po_file1);
#my ($hours, $minutes, $seconds) = ($time =~ m/(\d+):(\d+):(\d+)/);
my $version1 = $1 if ($po_url1 =~ m/(\d+\.\d+)/);
my $version2 = $1 if ($po_url2 =~ m/(\d+\.\d+)/);
$file_pies[-1]=~s/uk\-UA\-//i;
my $title = $file_pies[-1]." ($version1 «» $version2)";
&print_file($FILE,"<HTML>\n<head>\n<title>$title</title>\n<meta http-equiv=\"Content-Type\" content=\"text/html; charset=UTF-8\" />\n");

# td {white-space: pre; font-family: monospace; font-size:16px; spacing: 4px}
&print_file($FILE,"
<style type='text/css'>

    table {border-collapse: collapse}

    td {font-family: monospace; spacing: 4px}

    /*td:nth-child(2) {border-right: black solid 1px; padding-right: 10px}
    td:nth-child(3) {border-right: #999  solid 1px; padding-right: 10px;padding-left: 10px}
    td:nth-child(4) {padding-left: 10px}

    td:nth-child(1),
    td:nth-child(2) {color: #999}*/

   /*tr.disc_a td:nth-child(3),
   tr.disc_b td:nth-child(4),*/
    del {background-color: #fcc; color:#900;}
    ins {background-color: #cfc; color:#090;}


/* .change {background-color: #ddf;} */

    del, ins {font-weight: bold}

    ins {text-decoration: none} /* Remove annoying default underline of ins tag */

    h1 {font-family: sans-serif;}

  </style>
\n");

&print_file($FILE,"</head>\n<BODY>\n");

&print_file($FILE,"<h2><center>Порівняння $po_file1 та $po_file2</center></h2>\n");

&print_file($FILE,"<table width='100%'>\n");
&print_file($FILE,"<tr>\n");
&print_file($FILE,"<td width='50%'>Файл: $po_file1</td>\n");
&print_file($FILE,"<td width='50%'>Файл: $po_file2</td>\n");
&print_file($FILE,"</tr>\n");
&print_file($FILE,"<tr>\n");
&print_file($FILE,"<td width='50%'>".&po_info($po1)."</td>\n");
&print_file($FILE,"<td width='50%'>".&po_info($po2)."</td>\n");
&print_file($FILE,"</tr>\n");
&print_file($FILE,"</table>\n");

&print_file($FILE,"<br>\n");

#$po{'msgid_hash'}{$msgid_string}{'destination'} = $msgstr_string;
#$po{'msgid_hash'}{$msgid_string}{'translated'} = 1 unless exists $po{'flags'}[$phrase_number]{'fuzzy'};

my $msgid_both_count = 0;
my $msgid_uniq1_count = 0;
my $msgstr_translated_both_count = 0;
my $msgstr_nottranslated_both_count = 0;
#my $msgstr_translated1_count = 0;
#my $msgstr_translated2_count = 0;
#my $msgstr_nottranslated1_count = 0;
#my $msgstr_nottranslated2_count = 0;
#my $msgstr_fuzzy1_count = 0;
#my $msgstr_fuzzy2_count = 0;
my $msgstr_translated1eq2_count = 0;
my $msgstr_translated1noteq2_count = 0;
my %translated_both_not_equal;
#my $msgstr_translated2eq1_count = 0;
#my $msgstr_translated2noteq1_count = 0;
my $msgstr_translated1nottraslated2_count = 0;
my $msgstr_translated2nottraslated1_count = 0;
my %translated1nottraslated2;
my %translated2nottraslated1;
foreach my $msgid_string(sort keys %{$po1->{'msgid_hash'}})
   {
   
   #if(exists $po1->{'msgid_hash'}{$msgid_string}{'translated'})
   #      {$msgstr_translated1_count++;}
   #if(exists $po1->{'msgid_hash'}{$msgid_string}{'nottranslated'})
   #      {$msgstr_nottranslated1_count++;}
   #if(exists $po1->{'msgid_hash'}{$msgid_string}{'fuzzy'})
   #      {$msgstr_fuzzy1_count++;}
   
   if(exists $po2->{'msgid_hash'}{$msgid_string})
      {$msgid_both_count++;
      if(exists $po1->{'msgid_hash'}{$msgid_string}{'translated'})
         {
         if(exists $po2->{'msgid_hash'}{$msgid_string}{'translated'})   
            {#$msgstr_translated2_count++;
            $msgstr_translated_both_count++;
            if($po1->{'msgid_hash'}{$msgid_string}{'destination'} eq $po2->{'msgid_hash'}{$msgid_string}{'destination'})
               {$msgstr_translated1eq2_count++;}
            else
               {$msgstr_translated1noteq2_count++;
               $translated_both_not_equal{$msgid_string} = 1;               
               }
            }
         else
            {$msgstr_translated1nottraslated2_count++;
            $translated1nottraslated2{$msgid_string} = 1;  
            }
         }
      else
         {unless(exists $po2->{'msgid_hash'}{$msgid_string}{'translated'})   
            {$msgstr_nottranslated_both_count++;}         
         }   
         
      }
   else
      {$msgid_uniq1_count++;}
   
   }
my $msgid_uniq2_count = 0;
foreach my $msgid_string(sort keys %{$po2->{'msgid_hash'}})
   {
   #if(exists $po2->{'msgid_hash'}{$msgid_string}{'translated'})
   #   {$msgstr_translated2_count++;}
   #if(exists $po2->{'msgid_hash'}{$msgid_string}{'nottranslated'})
   #      {$msgstr_nottranslated2_count++;}
   #if(exists $po2->{'msgid_hash'}{$msgid_string}{'fuzzy'})
   #      {$msgstr_fuzzy2_count++;}
   
   if(exists $po1->{'msgid_hash'}{$msgid_string})
      {#$msgid_both_count++;
      if(exists $po2->{'msgid_hash'}{$msgid_string}{'translated'})
         {
         if(exists $po1->{'msgid_hash'}{$msgid_string}{'translated'})   
            {#$msgstr_translated2_count++;
            #$msgstr_translated_both_count++;
            if($po2->{'msgid_hash'}{$msgid_string}{'destination'} eq $po1->{'msgid_hash'}{$msgid_string}{'destination'})
               {#$msgstr_translated2eq1_count++;
               }
            else
               {#$msgstr_translated2noteq1_count++;
               }
            }
         else
            {$msgstr_translated2nottraslated1_count++;
            $translated2nottraslated1{$msgid_string} = 1; 
            }
         }

         
      }
   else
      {$msgid_uniq2_count++;}
   
   }
 
&print_file($FILE,"Спільних джерельних фраз = $msgid_both_count<br>\n");
&print_file($FILE,"Унікальних джерельних фраз у файлі '$po_file1' = $msgid_uniq1_count<br>\n");
&print_file($FILE,"Унікальних джерельних фраз у файлі '$po_file2' = $msgid_uniq2_count<br>\n");

&print_file($CSV,"$msgid_uniq1_count\t");
&print_file($CSV,"$msgid_uniq2_count\t");

&print_file($FILE,"Спільних перекладених фраз = $msgstr_translated_both_count<br>\n");
&print_file($FILE,"Спільних не перекладених (або неточно) фраз = $msgstr_nottranslated_both_count<br>\n");
#&print_file($FILE,"Перекладених фраз у файлі '$po_file1' = $msgstr_translated1_count<br>\n");
#&print_file($FILE,"Перекладених фраз у файлі '$po_file2' = $msgstr_translated2_count<br>\n");
#&print_file($FILE,"Не перекладених фраз у файлі '$po_file1' = $msgstr_nottranslated1_count<br>\n");
#&print_file($FILE,"Не перекладених фраз у файлі '$po_file2' = $msgstr_nottranslated2_count<br>\n");
#&print_file($FILE,"Не точно перекладених фраз у файлі '$po_file1' = $msgstr_fuzzy1_count<br>\n");
#&print_file($FILE,"Не точно перекладених фраз у файлі '$po_file2' = $msgstr_fuzzy2_count<br>\n");

&print_file($FILE,"Спільних перекладених фраз, що співпадають (1=2)     = $msgstr_translated1eq2_count<br>\n");
&print_file($FILE,"Спільних перекладених фраз, що не співпадають (1!=2) = $msgstr_translated1noteq2_count (<a href='#translated_not_eual'>перейти до списку</a>)<br>\n");

&print_file($CSV,"$msgstr_translated1noteq2_count\t");

#&print_file($FILE,"Спільних перекладених фраз, що співпадають (2=1) = $msgstr_translated2eq1_count<br>\n");
#&print_file($FILE,"Спільних перекладених фраз, що не співпадають (2!=1) = $msgstr_translated2noteq1_count<br>\n");

&print_file($FILE,"Перекладено у '$po_file1' але не перекладено у '$po_file2'  = $msgstr_translated1nottraslated2_count (<a href='#translated1_nottranslated2'>перейти до списку</a>)<br>\n");

&print_file($FILE,"Перекладено у '$po_file2' але не перекладено у '$po_file1'  = $msgstr_translated2nottraslated1_count (<a href='#translated2_nottranslated1'>перейти до списку</a>)<br>\n");

&print_file($CSV,"$msgstr_translated1nottraslated2_count\t");
&print_file($CSV,"$msgstr_translated2nottraslated1_count\n");

&print_file($FILE,"<hr>");
&print_file($FILE,"<a name='translated_not_eual'>Перелік спільних перекладених фраз, що не співпадають (усього $msgstr_translated1noteq2_count): <br>\n");
&print_file($FILE,"<table width='100%' border='1'>\n");
&print_file($FILE,"<tr>\n");
&print_file($FILE,"<td width='30%' align='center'> Джерельна фраза </td>\n");
&print_file($FILE,"<td width='5%' align='center'> 1</td>\n");
&print_file($FILE,"<td width='5%' align='center'> 2</td>\n");
&print_file($FILE,"<td width='60%' align='center'> Порівняння <del>$po_file1</del> та <ins>$po_file2</ins></td>\n");
&print_file($FILE,"</tr>\n");
foreach my $msgid_string(sort keys %translated_both_not_equal)
   {
   #my $po_url1  = $args[2];
   #my $po_url2  = $args[3];
   my $msgstr1 = $po1->{'msgid_hash'}{$msgid_string}{'destination'};
   my $msgstr2 = $po2->{'msgid_hash'}{$msgid_string}{'destination'};
   # https://translate.koha-community.org/uk/20.05/translate/uk-UA-staff-prog.po#search=%25s%20%25s%20%25s%20%25s%20%25s%20%25s%20%25s%20%25s%20%25s%20%25sCirculation%20%25sCatalog%20%25sPatrons%20%25sAcquisitions%20%25sAccounts%20%25sSerials%20%25s%20%25s%20%25s&sfields=source&soptions=exact
   my $url1_msgid = $po_url1."#search=".&uu_encode($msgid_string)."\&sfields=source\&soptions=exact,case";
   my $url2_msgid = $po_url2."#search=".&uu_encode($msgid_string)."\&sfields=source\&soptions=exact,case";
   my $url1_msgstr = $po_url1."#search=".&uu_encode($msgstr1)."\&sfields=target\&soptions=exact,case";
   my $url2_msgstr = $po_url2."#search=".&uu_encode($msgstr2)."\&sfields=target\&soptions=exact,case";
   #my $html = diff_strings($msgstr1,$msgstr2);
   my $html = diff_strings( { limit_onesided => 1 },$msgstr1,$msgstr2);
   &print_file($FILE,"<tr>\n");
   &print_file($FILE,"<td width='30%'>".&pre($msgid_string)."</td>\n");
   #&print_file($FILE,"<td width='25%'>".&pre($msgstr1)."</td>\n");
   &print_file($FILE,"<td width='5%' align='center'>1(<a href=\"$url1_msgid\">англ.</a>, <a href=\"$url1_msgstr\">укр.</a>)</td>\n");
   #&print_file($FILE,"<td width='25%'>".&pre($msgstr2)."</td>\n");
   &print_file($FILE,"<td width='5%' align='center'>2(<a href=\"$url2_msgid\">англ.</a>, <a href=\"$url2_msgstr\">укр.</a>)</td>\n");
   &print_file($FILE,"<td width='60%' bordercolor='#D3EDF6'>".$html."</td>\n");
   &print_file($FILE,"</tr>\n");   
   }
&print_file($FILE,"</table>\n");


&print_file($FILE,"<hr>");
&print_file($FILE,"<a name='translated1_nottranslated2'>Перелік перекладених фраз з файлу '$po_file1', які ще не перекладено у '$po_file2' (усього $msgstr_translated1nottraslated2_count): <br>\n");
&print_file($FILE,"<table  width='100%' border='1'>\n");
&print_file($FILE,"<tr>\n");
&print_file($FILE,"<td width='40%' align='center'> Джерельна фраза </td>\n");
&print_file($FILE,"<td width='5%' align='center'> 1</td>\n");
&print_file($FILE,"<td width='40%'> $po_file1</td>\n");
&print_file($FILE,"<td width='5%' align='center'> 2</td>\n");
&print_file($FILE,"<td width='10%'> $po_file2</td>\n");
&print_file($FILE,"</tr>\n");
foreach my $msgid_string(sort keys %translated1nottraslated2)
   {
   my $msgstr1 = $po1->{'msgid_hash'}{$msgid_string}{'destination'};
   my $url1_msgid = $po_url1."#search=".&uu_encode($msgid_string)."\&sfields=source\&soptions=exact,case";
   my $url2_msgid = $po_url2."#search=".&uu_encode($msgid_string)."\&sfields=source\&soptions=exact,case";
   my $url1_msgstr = $po_url1."#search=".&uu_encode($msgstr1)."\&sfields=target\&soptions=exact,case";
   &print_file($FILE,"<tr>\n");
   &print_file($FILE,"<td width='40%'>".&pre($msgid_string)."</td>\n");
   &print_file($FILE,"<td width='5%' align='center'>1(<a href=\"$url1_msgid\">англ.</a>, <a href=\"$url1_msgstr\">укр.</a>)</td>\n");
   &print_file($FILE,"<td width='40%'><ins>".&pre($msgstr1)."</ins></td>\n");
   &print_file($FILE,"<td width='5%' align='center'>2(<a href=\"$url2_msgid\">англ.</a>)</td>\n");
   &print_file($FILE,"<td width='10%'></td>\n");
   &print_file($FILE,"</tr>\n");
   }
&print_file($FILE,"</table>\n");


&print_file($FILE,"<hr>");
&print_file($FILE,"<a name='translated2_nottranslated1'>Перелік перекладених фраз з файлу '$po_file2', які ще не перекладено у '$po_file1' (усього $msgstr_translated2nottraslated1_count): <br>\n");
&print_file($FILE,"<table  width='100%' border='1'>\n");
&print_file($FILE,"<tr>\n");
&print_file($FILE,"<td width='40%' align='center'> Джерельна фраза </td>\n");
&print_file($FILE,"<td width='5%' align='center'> 1</td>\n");
&print_file($FILE,"<td width='40%'> $po_file1</td>\n");
&print_file($FILE,"<td width='5%' align='center'> 2</td>\n");
&print_file($FILE,"<td width='10%'> $po_file2</td>\n");
&print_file($FILE,"</tr>\n");
foreach my $msgid_string(sort keys %translated2nottraslated1)
   {
   my $msgstr2 = $po2->{'msgid_hash'}{$msgid_string}{'destination'};
   my $url1_msgid = $po_url1."#search=".&uu_encode($msgid_string)."\&sfields=source\&soptions=exact,case";
   my $url2_msgid = $po_url2."#search=".&uu_encode($msgid_string)."\&sfields=source\&soptions=exact,case";
   my $url2_msgstr = $po_url2."#search=".&uu_encode($msgstr2)."\&sfields=target\&soptions=exact,case";
   &print_file($FILE,"<tr>\n");
   &print_file($FILE,"<td width='40%'>".&pre($msgid_string)."</td>\n");
   &print_file($FILE,"<td width='5%' align='center'>1(<a href=\"$url1_msgid\">англ.</a>)</td>\n");
   &print_file($FILE,"<td width='40%'></td>\n");
   &print_file($FILE,"<td width='5%' align='center'>2(<a href=\"$url2_msgid\">англ.</a>, <a href=\"$url2_msgstr\">укр.</a>)</td>\n");
   &print_file($FILE,"<td width='10%'><ins>".&pre($msgstr2)."</ins></td>\n");
   &print_file($FILE,"</tr>\n");   
   }
&print_file($FILE,"</table>\n");
	
	
&print_file($FILE,"</BODY>\n</HTML>");


close($FILE);
close($CSV);








# ================================== ROUTINES ===========================================

# //////// &print_file($FILE,$string) //////////////
sub print_file($$)
	{
	my $FILE 	= shift;
	my $string = shift; # return unless defined $string;
	print $FILE &encode($encoding_output_file,$string);
	}
# \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

# //////// &print_terminal($string) //////////////
sub print_terminal($)
	{my $string = shift; # return unless defined $string;
	print &encode($encoding_terminal_output,$string);
	}
# \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

# //////// &po_parse($po_file) ///////////////////
sub po_parse($)
   {my $po_file = shift;
   my %po;
   my $po_aref = Locale::PO->load_file_asarray($po_file);
   # http://translate.koha-community.org/uk/36/uk-UA-i-opac-t-prog-v-3006000.po/translate/?page=219
# 2182 - http://translate.koha-community.org/uk/36/uk-UA-i-opac-t-prog-v-3006000.po/translate/?unit=2455413
   #. %1$s:  END 
   #. %2$s:  END 
   #. %3$s:  END 
   #. %4$s:  END 
   #. %5$s:  UNLESS ( loop.last || languages_loo.plural ) 
   #. %6$s:  END 
   #: opac-tmpl/prog/en/includes/opac-bottom.inc:37
   #, fuzzy, c-format
   #msgid "%s %s %s %s%s | %s "
   #msgstr "%s %s %s %s %s %s %s "
   #my %po;
   my $phrase_number = -1;
   #my @msgid1; # $msgid[$phrase_number] = $msgid_string;
   #my @msgstr1; # $msgstr[$phrase_number] = $msgstr_string;
   #my @flags1; # ex. $flag[$phrase_number]{fuzzy} = 1
   #my %path1; # $path{$path_string}{$line_number}[$occ]
   #my %path_strings1; # $path_strings{$path_string}
   #my %path_bytes1; # $path_bytes{$path_string}
   #my %path_untranslated_strings1; # $path_translated_strings{$path_string}
   #my %path_untranslated_bytes1; # $path_translated_bytes{$path_string}
   #my %automatics1; # $automatics{$phrase_number}[$d] = $expl;
   
   foreach my $po_item(@{$po_aref})
   {$phrase_number++;
   
   &print_terminal($d->get("\n-------------------------- ITEM $phrase_number:\n"));
   
   # MSGID: Item Type:
   # MSGSTR: Тип одиниці: 
   # msgid, msgstr
	my $msgid_string  = &decode($encoding_input_file,$po_item->dequote($po_item->msgid()));
	#next if length($msgid_string) == 0;
	my $msgstr_string = &decode($encoding_input_file,$po_item->dequote($po_item->msgstr()));
	unless(defined $msgstr_string)
	   {
	   my %plural_hash = %{$po_item->msgstr_n()};
	   my %plural_dequote;
	   foreach my $plural_i(keys %plural_hash)
	      {
	      $plural_dequote{$plural_i} = &decode($encoding_input_file,$po_item->dequote($plural_hash{$plural_i}));
	      }
	   $msgstr_string = join(" ",sort values %plural_dequote);
	   }
	&print_terminal("MSGID[$phrase_number]: '$msgid_string'\n");
	&print_terminal("MSGSTR[$phrase_number]: '$msgstr_string'\n");
	$po{'msgid'}[$phrase_number] = $msgid_string;
   $po{'msgstr'}[$phrase_number] = $msgstr_string;
	
	if(exists $po{'msgid_hash'}{$msgid_string})
	   {&print_terminal("ERROR: phrase:\n$msgid_string\ndublets in row $phrase_number\n");
	   #exit;
	   }
	
	next if $po_item->obsolete();
	
	$po{'msgid_hash'}{$msgid_string}{'destination'} = $msgstr_string;
	$po{'msgid_hash'}{$msgid_string}{'number'}      = $phrase_number;
	
	
	# fuzzy, c_format, php_format, python-format, no-c-format
	if(exists $po_item->{'_flags'})
		{my $flag_count = 0;
		foreach my $flag(@{$po_item->{'_flags'}})
			{my $decoded_flag = &decode($encoding_input_file,$flag);
			&print_terminal($d->get("FLAG: ")."'$decoded_flag'\n");
			$po{'flags'}[$phrase_number]{$decoded_flag} = 1;
			$flag_count++;
			}
		}
	
	if(exists $po{'flags'}[$phrase_number]{'fuzzy'})
	   {$po{'msgid_hash'}{$msgid_string}{'fuzzy'} = 1;}
	elsif(length($msgstr_string) == 0)
	   {$po{'msgid_hash'}{$msgid_string}{'nottranslated'} = 1;}
	else
	   {$po{'msgid_hash'}{$msgid_string}{'translated'} = 1;}
		
   # opac-tmpl/prog/en/includes/opac-facets.inc:6
   # opac-tmpl/prog/en/includes/opac-facets.inc:20
   # opac-tmpl/prog/en/includes/opac-facets.inc:21
   # reference
	if($po_item->reference())
		{my $reference = &decode($encoding_input_file,$po_item->reference());
		&print_terminal($d->get("REFERENCE: ")."$reference\n");
		my @lines = split(/[\n\s]/,$reference);
		foreach my $line(@lines)
		   {(my $path_string, my $line_number) = split(/\:/,$line,2);
		   &print_terminal("\$path_string='$path_string', \$line_number='$line_number'\n");
		   unless($line_number=~/^\d+$/){&print_terminal($d->get("WARNING: ")."\$line_number='$line_number' ".$d->get("is not number")."\n");};
		   if(exists $po{'path'}{$path_string}{$line_number})
		      {my $occ = scalar(@{$po{'path'}{$path_string}{$line_number}});
		      $po{'path'}{$path_string}{$line_number}[$occ] = $phrase_number;}
		   else
		      {$po{'path'}{$path_string}{$line_number}[0] = $phrase_number;}
		   
		   
		   if(exists $po{'path_strings'}{$path_string})
		      {$po{'path_strings'}{$path_string}++;
		      $po{'path_bytes'}{$path_string} += length($po{'msgid'}[$phrase_number]);
		      }
		   else
		      {$po{'path_strings'}{$path_string} = 1;
		      $po{'path_bytes'}{$path_string} = length($po{'msgid'}[$phrase_number]);
		      }
		   
		   if((exists $po{'flags'}[$phrase_number]{'fuzzy'}) or (length($po{'msgstr'}[$phrase_number]) == 0))
		      {
		      if(exists $po{'path_untranslated_strings'}{$path_string})
		         {$po{'path_untranslated_strings'}{$path_string}++;
		         $po{'path_untranslated_bytes'}{$path_string} += length($po{'msgid'}[$phrase_number]);
		         }
		      else
		         {$po{'path_untranslated_strings'}{$path_string} = 1;
		         $po{'path_untranslated_bytes'}{$path_string} = length($po{'msgid'}[$phrase_number]);
		         }
		      }
	      
		   }
		}
   
   # comment
	if($po_item->comment())
		{
		my $comment = &decode($encoding_input_file,$po_item->comment());
		&print_terminal($d->get("COMMENT: ")."'$comment'\n");
		}
	
	# msgctxt
	if($po_item->msgctxt())
		{
		my $dequote_msgctxt = $po_item->dequote(&decode($encoding_input_file,$po_item->msgctxt()));
		&print_terminal("MSGCTXT: '$dequote_msgctxt'\n");
		}
   
   # automatic
	if($po_item->automatic())
		{
		my $automatic = &decode($encoding_input_file,$po_item->automatic());
		&print_terminal($d->get("AUTOMATIC: ")."'$automatic'\n");
		# %2$s:  IF ( facets_loo.expandable ) 
		# %2$s:  END 
		my @automatic_lines = split(/\n/,$automatic);
		foreach my $line(@automatic_lines)
		   {#&print_terminal("\$line='$line'\n");
		   if($line=~/^\%(\d+)\$s\:/)
		      {
		      my $d = $1;
		      my $expl = $line;
		      $expl=~s/^\%(\d+)\$s\:\s+//;
		      $po{'automatics'}{$phrase_number}[$d] = $expl;
		      &print_terminal("\$automatics{$phrase_number}[$d]='$expl'\n");
		      }
		   }
		}
   
   }
   
   return \%po;
   }
# \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

# //////// &po_info(\%po) //////////////
sub po_info($)
	{
	my $po 	= shift;
   my $s = '';
   #return $s if (exists $po->{'msgid'} and ref(@{$po->{'msgid'}}) ne 'ARRAY');
   return $s if not defined $po;
   return $s if not defined $po->{'msgid'};
   return $s if ref(@{$po->{'msgid'}}) ne 'ARRAY';
   #   {
   #   &print_terminal("\nDUMP(\$po)=",&Dumper($po),"\n");      
   #   }
   my $string_number = scalar(@{$po->{'msgid'}});
   $s.="Кількість рядків = $string_number<br>";
   
   my %flags_counts;
   my $fuzzy_count = 0;
   my $translated_count = 0;
   my $nottranslated_count = 0;
   foreach my $i(0..scalar(@{$po->{'flags'}})-1)    
      {foreach my $flag(keys %{$po->{'flags'}[$i]})
         {$flags_counts{$flag}++ if $po->{'flags'}[$i]{$flag} == 1;}      
      }
   foreach my $msgid_string(keys %{$po->{'msgid_hash'}})
      {$fuzzy_count++         if exists $po->{'msgid_hash'}{$msgid_string}{'fuzzy'};
   	$nottranslated_count++ if exists $po->{'msgid_hash'}{$msgid_string}{'nottranslated'};
   	$translated_count++    if exists $po->{'msgid_hash'}{$msgid_string}{'translated'};
      }
   $s.="Прапорці = ".join(" | ",sort keys %flags_counts)."<br>";
   $s.="Перекладено рядків = $translated_count (".sprintf("%.2f",($translated_count*100/$string_number))." \%)<br>";
   
   &print_file($CSV,"$translated_count\t");
   &print_file($CSV,sprintf("%.2f",($translated_count*100/$string_number))."\t");
   &print_file($CSV,($fuzzy_count+$nottranslated_count)."\t");
   &print_file($CSV,sprintf("%.2f",(($nottranslated_count+$fuzzy_count)*100/$string_number))."\t");
   
   $s.="Не перекладено рядків (усього) = ".($fuzzy_count+$nottranslated_count)." (".sprintf("%.2f",(($nottranslated_count+$fuzzy_count)*100/$string_number))." \%)<br>";
   $s.="Не перекладено рядків (без врахування неточних) = $nottranslated_count (".sprintf("%.2f",($nottranslated_count*100/$string_number))." \%)<br>";
   $s.="Не точно перекладено рядків = $fuzzy_count (".sprintf("%.2f",($fuzzy_count*100/$string_number))." \%)<br>";
   
   return $s;
	}
# \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

# //////// &pre($string) //////////////
sub pre($)
	{
	my $string 	= shift;
	$string=~s/\&/&amp;/g;
	$string=~s/\</&lt;/g;
	$string=~s/\>/&gt;/g;
	return $string;
	}
# \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

# //////// &uu_encode($string) //////////////
sub uu_encode($)
	{
	my $string 	= shift;
	$string=~s/\s+$//g;
	$string=~s/\%/\%25/g;
	$string=~s/\&/\%26/g;
	$string=~s/\;/\%3B/g;
	$string=~s/\+/\%2B/g;
	$string=~s/\\/\%5C/g;
	$string=~s/\//\%2F/g;
	$string=~s/\#/\%23/g;
	$string=~s/\"/\%22/g;
	return $string;
	}
# \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

